#include <iostream>
#include <algorithm>
#include <set>
#include <Eigen/Dense>

#include <sensor_msgs/image_encodings.h>
#include <random_numbers/random_numbers.h>

#include <msckf_vio/CameraMeasurement.h>
#include <msckf_vio/TrackingInfo.h>
#include <msckf_vio/image_processor.h>

namespace msckf_vio {
namespace image_handler {


cv::Point2f pinholeDownProject(const cv::Point2f& p_in, const cv::Vec4d& intrinsics)
{
  return cv::Point2f(p_in.x * intrinsics[0] + intrinsics[2], p_in.y * intrinsics[1] + intrinsics[3]);
}

cv::Point2f pinholeUpProject(const cv::Point2f& p_in, const cv::Vec4d& intrinsics)
{
  return cv::Point2f((p_in.x - intrinsics[2])/intrinsics[0], (p_in.y - intrinsics[3])/intrinsics[1]);
}

void undistortImage(
  cv::InputArray src,
  cv::OutputArray dst,
  const std::string& distortion_model,
  const cv::Vec4d& intrinsics,
  const cv::Vec4d& distortion_coeffs)
{

  const cv::Matx33d K(intrinsics[0], 0.0, intrinsics[2],
                      0.0, intrinsics[1], intrinsics[3],
                      0.0, 0.0, 1.0);

  if (distortion_model == "pre-equidistant")
      cv::fisheye::undistortImage(src, dst, K, distortion_coeffs, K);
  else if (distortion_model == "equidistant")
      src.copyTo(dst);
  else if (distortion_model == "pre-radtan")
      cv::undistort(src, dst, K, distortion_coeffs, K);
  else if (distortion_model == "radtan")
      src.copyTo(dst);
}

void undistortPoint(
    const cv::Point2f& pt_in,
    const cv::Vec4d& intrinsics,
    const std::string& distortion_model,
    const cv::Vec4d& distortion_coeffs,
    cv::Point2f& pt_out,
    const cv::Matx33d &rectification_matrix,
    const cv::Vec4d &new_intrinsics) {


  std::vector<cv::Point2f> pts_in;
  std::vector<cv::Point2f> pts_out;
  pts_in.push_back(pt_in);
  if (pts_in.size() == 0) return;

  const cv::Matx33d K(
      intrinsics[0], 0.0, intrinsics[2],
      0.0, intrinsics[1], intrinsics[3],
      0.0, 0.0, 1.0);

  const cv::Matx33d K_new(
      new_intrinsics[0], 0.0, new_intrinsics[2],
      0.0, new_intrinsics[1], new_intrinsics[3],
      0.0, 0.0, 1.0);

  if (distortion_model == "radtan") {
    cv::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                        rectification_matrix, K_new);
  } 
  // equidistant
  else if (distortion_model == "equidistant") {
    cv::fisheye::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                                 rectification_matrix, K_new);
  }
  // fov
  else if (distortion_model == "fov") {
    for(int i = 0; i < pts_in.size(); i++)
    {
      float omega = distortion_coeffs[0];
      float rd = sqrt(pts_in[i].x * pts_in[i].x + pts_in[i].y * pts_in[i].y);
      float ru = tan(rd * omega)/(2 * tan(omega / 2));

      cv::Point2f newPoint(
              ((pts_in[i].x - intrinsics[2]) / intrinsics[0]) * (ru / rd),
              ((pts_in[i].y - intrinsics[3]) / intrinsics[1]) * (ru / rd));

      pts_out.push_back(newPoint);
    }
   }
  else if (distortion_model == "pre-equidistant" or distortion_model == "pre-radtan")
  {
    std::vector<cv::Point2f> temp_pts_out;
    for(int i = 0; i < pts_in.size(); i++)
      temp_pts_out.push_back(pinholeUpProject(pts_in[i], intrinsics));

    pts_out = temp_pts_out;
  } 
   else {
    ROS_WARN_ONCE("The model %s is unrecognized, use radtan instead...",
                  distortion_model.c_str());
    cv::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                        rectification_matrix, K_new);
  }
  pt_out = pts_out[0];
  return;
}

void undistortPoints(
    const std::vector<cv::Point2f>& pts_in,
    const cv::Vec4d& intrinsics,
    const std::string& distortion_model,
    const cv::Vec4d& distortion_coeffs,
    std::vector<cv::Point2f>& pts_out,
    const cv::Matx33d &rectification_matrix,
    const cv::Vec4d &new_intrinsics) {

  if (pts_in.size() == 0) return;

  const cv::Matx33d K(
      intrinsics[0], 0.0, intrinsics[2],
      0.0, intrinsics[1], intrinsics[3],
      0.0, 0.0, 1.0);

  const cv::Matx33d K_new(
      new_intrinsics[0], 0.0, new_intrinsics[2],
      0.0, new_intrinsics[1], new_intrinsics[3],
      0.0, 0.0, 1.0);

  if (distortion_model == "radtan") {
    cv::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                        rectification_matrix, K_new);
  } 
  else if (distortion_model == "equidistant") {
    cv::fisheye::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                                 rectification_matrix, K_new);
  } 
  else if (distortion_model == "fov") {
    for(int i = 0; i < pts_in.size(); i++)
    {
      float omega = distortion_coeffs[0];
      float rd = sqrt(pts_in[i].x * pts_in[i].x + pts_in[i].y * pts_in[i].y);
      float ru = tan(rd * omega)/(2 * tan(omega / 2));

      cv::Point2f newPoint(
              ((pts_in[i].x - intrinsics[2]) / intrinsics[0]) * (ru / rd),
              ((pts_in[i].y - intrinsics[3]) / intrinsics[1]) * (ru / rd));

      pts_out.push_back(newPoint);
    }
  }
  else if (distortion_model == "pre-equidistant" or distortion_model == "pre-radtan")
  {
    std::vector<cv::Point2f> temp_pts_out;
    for(int i = 0; i < pts_in.size(); i++)
      temp_pts_out.push_back(pinholeUpProject(pts_in[i], intrinsics));

    pts_out = temp_pts_out;

  }  
  else {
    ROS_WARN_ONCE("The model %s is unrecognized, use radtan instead...",
                  distortion_model.c_str());
    cv::undistortPoints(pts_in, pts_out, K, distortion_coeffs,
                        rectification_matrix, K_new);
  }

  return;
}

std::vector<cv::Point2f> distortPoints(
    const std::vector<cv::Point2f>& pts_in,
    const cv::Vec4d& intrinsics,
    const std::string& distortion_model,
    const cv::Vec4d& distortion_coeffs) {

  const cv::Matx33d K(intrinsics[0], 0.0, intrinsics[2],
                      0.0, intrinsics[1], intrinsics[3],
                      0.0, 0.0, 1.0);

  std::vector<cv::Point2f> pts_out;
  if (distortion_model == "radtan") {
    std::vector<cv::Point3f> homogenous_pts;
    cv::convertPointsToHomogeneous(pts_in, homogenous_pts);
    cv::projectPoints(homogenous_pts, cv::Vec3d::zeros(), cv::Vec3d::zeros(), K,
                      distortion_coeffs, pts_out);
  } else if (distortion_model == "equidistant") {
    cv::fisheye::distortPoints(pts_in, pts_out, K, distortion_coeffs);
  }
  else if (distortion_model == "fov") {
    for(int i = 0; i < pts_in.size(); i++)
    {
      // based on 'straight lines have to be straight'
      float ru = sqrt(pts_in[i].x * pts_in[i].x + pts_in[i].y * pts_in[i].y);
      float omega = distortion_coeffs[0];
      float rd = 1 / (omega)*atan(2*ru*tan(omega / 2));
      
      cv::Point2f newPoint(
              pts_in[i].x * (rd/ru) * intrinsics[0] + intrinsics[2],
              pts_in[i].y * (rd/ru) * intrinsics[1] + intrinsics[3]);

      pts_out.push_back(newPoint);
    }
  }
  else if (distortion_model == "pre-equidistant" or distortion_model == "pre-radtan")
  {
    std::vector<cv::Point2f> temp_pts_out;
    for(int i = 0; i < pts_in.size(); i++)
      temp_pts_out.push_back(pinholeDownProject(pts_in[i], intrinsics));

    pts_out = temp_pts_out;
  }   
   else {
    ROS_WARN_ONCE("The model %s is unrecognized, using radtan instead...",
                  distortion_model.c_str());
    std::vector<cv::Point3f> homogenous_pts;
    cv::convertPointsToHomogeneous(pts_in, homogenous_pts);
    cv::projectPoints(homogenous_pts, cv::Vec3d::zeros(), cv::Vec3d::zeros(), K,
                      distortion_coeffs, pts_out);
  }

  return pts_out;
}

cv::Point2f distortPoint(
    const cv::Point2f& pt_in,
    const cv::Vec4d& intrinsics,
    const std::string& distortion_model,
    const cv::Vec4d& distortion_coeffs) {

  std::vector<cv::Point2f> pts_in;
  pts_in.push_back(pt_in);

  const cv::Matx33d K(intrinsics[0], 0.0, intrinsics[2],
                      0.0, intrinsics[1], intrinsics[3],
                      0.0, 0.0, 1.0);

  std::vector<cv::Point2f> pts_out;
  if (distortion_model == "radtan") {
    std::vector<cv::Point3f> homogenous_pts;
    cv::convertPointsToHomogeneous(pts_in, homogenous_pts);
    cv::projectPoints(homogenous_pts, cv::Vec3d::zeros(), cv::Vec3d::zeros(), K,
                      distortion_coeffs, pts_out);
  } else if (distortion_model == "equidistant") {
    cv::fisheye::distortPoints(pts_in, pts_out, K, distortion_coeffs);
  }
  else if (distortion_model == "fov") {
    for(int i = 0; i < pts_in.size(); i++)
    {
      // based on 'straight lines have to be straight'
      float ru = sqrt(pts_in[i].x * pts_in[i].x + pts_in[i].y * pts_in[i].y);
      float omega = distortion_coeffs[0];
      float rd = 1 / (omega)*atan(2*ru*tan(omega / 2));
      
      cv::Point2f newPoint(
              pts_in[i].x * (rd/ru) * intrinsics[0] + intrinsics[2],
              pts_in[i].y * (rd/ru) * intrinsics[1] + intrinsics[3]);

      pts_out.push_back(newPoint);
    }
  }
  else if (distortion_model == "pre-equidistant" or distortion_model == "pre-radtan")
  {
    std::vector<cv::Point2f> temp_pts_out;
    for(int i = 0; i < pts_in.size(); i++)
      pts_out.push_back(pinholeDownProject(pts_in[i], intrinsics));

    pts_out = temp_pts_out;
  }    
   else {
    ROS_WARN_ONCE("The model %s is unrecognized, using radtan instead...",
                  distortion_model.c_str());
    std::vector<cv::Point3f> homogenous_pts;
    cv::convertPointsToHomogeneous(pts_in, homogenous_pts);
    cv::projectPoints(homogenous_pts, cv::Vec3d::zeros(), cv::Vec3d::zeros(), K,
                      distortion_coeffs, pts_out);
  }

  return pts_out[0];
}

  } // namespace image_handler
} // namespace msckf_vio
